#include <linux/compat.h>
#include "drmP.h"
#include "drm.h"
#include "via_chrome9_drm.h"
#include "via_chrome9_dma.h"
#include "via_chrome9_drv.h"

struct drm_via_chrome9_init32{
	u32 func;
	int chip_agp;
	int chip_index;
	int chip_sub_index;
	int usec_timeout;
	unsigned int   sarea_priv_offset;
	unsigned int   fb_cpp;
	unsigned int   front_offset;
	unsigned int   back_offset;
	unsigned int   depth_offset;
	u32	mmio_handle;
	u32	dma_handle;
	u32	fb_handle;
	unsigned int   front_handle;
	unsigned int   back_handle;
	unsigned int   depth_handle;

	unsigned int   fb_tex_offset;
	unsigned int   fb_tex_size;

	unsigned int   agp_tex_size;
	u32	agp_tex_handle;
	unsigned int   shadow_size;
	u32	shadow_handle;
	unsigned int   garttable_size;
	unsigned int   garttable_offset;
	u32	available_fb_size;
	u32	fb_base_address;
	u32	DMA_size;
	u32	DMA_phys_address;
	u32	agp_type;
	u32	hostBlt_handle;
};

static int compat_via_chrome9_init(struct file *filp, unsigned int cmd,
		unsigned long arg)
{
	struct drm_via_chrome9_init32 chrome9_init32;
	struct drm_via_chrome9_init __user *init;
	int err;

	if (copy_from_user(&chrome9_init32, (void  __user *)arg,
		sizeof(struct  drm_via_chrome9_init32)))
		return -EFAULT;
	init = compat_alloc_user_space(sizeof(*init));

	if (!access_ok(VERIFY_WRITE, init, sizeof(*init))
		|| __put_user(chrome9_init32.mmio_handle, &init->mmio_handle)
		|| __put_user(chrome9_init32.dma_handle, &init->dma_handle)
		|| __put_user(chrome9_init32.fb_handle, &init->fb_handle)
		|| __put_user(chrome9_init32.agp_tex_handle,
			&init->agp_tex_handle)
		|| __put_user(chrome9_init32.shadow_handle,
			&init->shadow_handle)
		|| __put_user(chrome9_init32.available_fb_size,
			&init->available_fb_size)
		|| __put_user(chrome9_init32.fb_base_address,
			&init->fb_base_address)
		|| __put_user(chrome9_init32.DMA_size, &init->DMA_size)
		|| __put_user(chrome9_init32.DMA_phys_address,
			&init->DMA_phys_address)
		|| __put_user(chrome9_init32.hostBlt_handle,
			&init->hostBlt_handle)
		|| __put_user(chrome9_init32.garttable_offset,
			&init->garttable_offset)
		|| __put_user(chrome9_init32.garttable_size,
			&init->garttable_size)
		|| __put_user(chrome9_init32.shadow_size, &init->shadow_size)
		|| __put_user(chrome9_init32.agp_tex_size, &init->agp_tex_size)
		|| __put_user(chrome9_init32.fb_tex_size, &init->fb_tex_size)
		|| __put_user(chrome9_init32.fb_tex_offset,
				&init->fb_tex_offset)
		|| __put_user(chrome9_init32.front_handle, &init->front_handle)
		|| __put_user(chrome9_init32.back_offset, &init->back_offset)
		|| __put_user(chrome9_init32.sarea_priv_offset,
				&init->sarea_priv_offset)
		|| __put_user(chrome9_init32.usec_timeout, &init->usec_timeout)
		|| __put_user(chrome9_init32.chip_sub_index,
				&init->chip_sub_index)
		|| __put_user(chrome9_init32.chip_index, &init->chip_index)
		|| __put_user(chrome9_init32.chip_agp, &init->chip_agp)
		|| __put_user(chrome9_init32.agp_type, &init->agp_type)
		|| __put_user(chrome9_init32.func, &init->func))
			return -EFAULT;

#if CHROME9_COMPAT_DEBUG
	printk(KERN_ALERT "CHROME9 COMPAT_DEBUG:""init->func=%d\n",
		chrome9_init32.func);
#endif
	err = drm_ioctl(filp->f_path.dentry->d_inode, filp,
		DRM_IOCTL_VIA_CHROME9_INIT, (unsigned long)init);
#if CHROME9_COMPAT_DEBUG
	if (!err)
		printk(KERN_ALERT "CHROME9_COMPAT""compat_via_chrome9_init"
			" called! ""drm_ioctl return success!\n");
	else{
		printk(KERN_ALERT "CHROME9_COMPAT_ERROR""compat_via_chrome9_init"
			" called!drm_ioctl return failed ! err=0x%x\n",err);
		return err;
	}
#endif
	return 0;
}

struct drm_via_chrome9_flush32 {
	u32	dma_cmd_type;
	int    cmd_idx;
	int    cmd_offset;
	int    cmd_size;
	u32    dma_kickoff[2];
	u32	usermode_dma_buf;
};
static int compat_via_chrome9_ioctl_flush(struct file *filp, unsigned int cmd,
	unsigned long arg)
{
	struct drm_via_chrome9_flush32 dma_info32;
	struct drm_via_chrome9_flush __user *dma_info;
	int err;

	if (copy_from_user(&dma_info32, (void __user *)arg, sizeof(dma_info32)))
		return -EFAULT;

#if CHROME9_COMPAT_DEBUG
	printk(KERN_ALERT "CHROME9 COMPAT_DEBUG:"" compat_ioctl_flush called! \n");
#endif
	dma_info  = compat_alloc_user_space(sizeof(*dma_info));

	if (!access_ok(VERIFY_WRITE, dma_info, sizeof(*dma_info))
		|| __put_user(dma_info32.dma_cmd_type, &dma_info->dma_cmd_type)
		|| __put_user(dma_info32.cmd_size, &dma_info->cmd_size)
		|| __put_user(dma_info32.usermode_dma_buf,
			(unsigned long *)&dma_info->usermode_dma_buf))
		return -EFAULT;

	err = drm_ioctl(filp->f_path.dentry->d_inode, filp,
		DRM_IOCTL_VIA_CHROME9_FLUSH, (unsigned long)dma_info);
	if (err)
		return err;
	return 0;
}

struct drm_via_chrome9_memory_alloc32 {
	u32 heap_type;
	struct {
		void *lpL1Node;
		unsigned int       alcL1Tag;
		unsigned int       usageCount;
		unsigned int       dwVersion;
		unsigned int       dwResHandle;
		unsigned int       dwProcessID;
	} heap_info;
	unsigned int flags;
	unsigned int size;
	u32 physaddress;
	u32 offset;
	unsigned int align;
	u32 linearaddress;
};

static int compat_via_chrome9_ioctl_allocate_mem_wrapper(struct file *filp,
		unsigned int cmd, unsigned long arg)
{

	struct drm_via_chrome9_memory_alloc32 memory_alloc32;
	struct drm_via_chrome9_memory_alloc *memory_alloc;
	int err;

	if (copy_from_user(&memory_alloc32, (void  __user *)arg,
				sizeof(struct drm_via_chrome9_memory_alloc32)))
		return -EFAULT;

	memory_alloc = compat_alloc_user_space(sizeof(*memory_alloc));

	if (!access_ok(VERIFY_WRITE, memory_alloc, sizeof(*memory_alloc))
		|| __put_user(memory_alloc32.flags, &memory_alloc->flags)
		|| __put_user(memory_alloc32.align, &memory_alloc->align)
		|| __put_user(memory_alloc32.heap_type,
			&memory_alloc->heap_type)
		|| __put_user(memory_alloc32.size, &memory_alloc->size))
		return -EFAULT;


	err = drm_ioctl(filp->f_path.dentry->d_inode, filp,
		DRM_IOCTL_VIA_CHROME9_ALLOCATE_VIDEO_MEM,
		(unsigned long)memory_alloc);
	if (err) {
		printk(KERN_ALERT "[CHROME9_COMPAT_ERROR]"
			"compat_via_chrome9_ioctl_allocate_mem_wrapper"
			"  drm_ioctl failed!! \n");
		return err;
	}
	/*copy all terms which may changed to memory_alloc32,
	  except linear_address, this becasue this value
	  can't be accessed in userspace*/
	if (__get_user(memory_alloc32.physaddress, &memory_alloc->physaddress)
		|| __get_user(memory_alloc32.offset, &memory_alloc->offset)
		|| __get_user(memory_alloc32.size, &memory_alloc->size)
		|| __get_user(memory_alloc32.heap_info.lpL1Node,
			&memory_alloc->heap_info.lpL1Node))
		return -EFAULT;
	if (copy_to_user((void __user *)arg, &memory_alloc32,
				sizeof(memory_alloc32)))
		return -EFAULT;
#if CHROME9_COMPAT_DEBUG
	printk(KERN_ALERT "CHROME9_COMPAT" "compat_via_chrome9_ioctl_allocate"
		"_mem_wrapper called successfully \n");
#endif
	return 0;

}

static int compat_via_chrome9_ioctl_free_mem_wrapper(struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	struct drm_via_chrome9_memory_alloc32 memory_alloc32;
	struct drm_via_chrome9_memory_alloc __user *memory_alloc;
	int err;

	if (copy_from_user(&memory_alloc32, (void  __user *)arg,
			sizeof(struct drm_via_chrome9_memory_alloc32)))
		return -EFAULT;

	memory_alloc = compat_alloc_user_space(sizeof(*memory_alloc));

	if (!access_ok(VERIFY_WRITE, memory_alloc, sizeof(*memory_alloc))
		|| __put_user(memory_alloc32.heap_info.lpL1Node,
			&memory_alloc->heap_info.lpL1Node))
		return -EFAULT;

	err = drm_ioctl(filp->f_path.dentry->d_inode, filp,
		DRM_IOCTL_VIA_CHROME9_FREE_VIDEO_MEM,
		(unsigned long)memory_alloc);
	return err;
}

struct drm_via_chrome9_mem32 {
	uint32_t context;
	uint32_t type;
	u32 size;
	u32 index;
	u32 offset;
};

static int compat_via_chrome9_ioctl_allocate_mem_base(struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	struct drm_via_chrome9_mem32 mem32;
	struct drm_via_chrome9_mem *mem;
	int err;

	if (copy_from_user(&mem32, (void __user *)arg, sizeof(mem32)))
		return -EFAULT;

	mem = compat_alloc_user_space(sizeof(*mem));

	if (!access_ok(VERIFY_WRITE, mem, sizeof(*mem))
		|| __put_user(mem32.size, &mem->size)
		|| __put_user(mem32.type, &mem->type))
		return -EFAULT;
#if CHROME9_COMPAT_DEBUG
	printk(KERN_ALERT "[CHROME9_COMPAT_DUBUG]"" alloc mem_base request "
		"type=0x%x! ,drm_ioctl copy size=0x%x \n", mem32.type,
			_IOC_SIZE(DRM_IOCTL_VIA_CHROME9_ALLOCMEM));
#endif
	err = drm_ioctl(filp->f_path.dentry->d_inode, filp,
			DRM_IOCTL_VIA_CHROME9_ALLOCMEM, (unsigned long)mem);
	if (err)
		return err;
#if CHROME9_COMPAT_DEBUG
	printk(KERN_ALERT "[CHROME9_COMPAT_DEBUG]"" dump "
		"struct drm_via_chrome9_mem in user stack: \n");
	u32 *pmem = mem;
	int i = 0;
	for(i = 0; i < sizeof(*mem)/4; i++)
		printk(" 0x%x \n",pmem[i]);
#endif

	if (__get_user(mem32.size, &mem->size)
		|| __get_user(mem32.offset, &mem->offset)
		|| __get_user(mem32.index, &mem->index))
		return -EFAULT;
#if CHROME9_COMPAT_DEBUG
	printk(KERN_ALERT "sheldon compat"" alloc mem_base request success! "
			"offset=0x%x index=0x%x \n", mem32.offset, mem32.index);
#endif
	if (copy_to_user((void __user *)arg, &mem32, sizeof(mem32)))
		return -EFAULT;

		return 0;
}

static int compat_via_chrome9_ioctl_freemem_base(struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	struct drm_via_chrome9_mem32 mem32;
	struct drm_via_chrome9_mem *mem;
	int err;

	if (copy_from_user(&mem32, (void __user *)arg, sizeof(mem32)))
		return -EFAULT;

	mem = compat_alloc_user_space(sizeof(*mem));

	if (!access_ok(VERIFY_WRITE, mem, sizeof(*mem))
		|| __put_user(mem32.index, &mem->index))
		return -EFAULT;

	err = drm_ioctl(filp->f_path.dentry->d_inode, filp,
			DRM_IOCTL_VIA_CHROME9_FREEMEM, (unsigned long)mem);
	if (err)
		return err;
	return 0;
}


struct umd_branch_buf_rec32 {
	u32 id;
	u32 status;
	u32 offset;         /* in unit of byte */
	u32 size;           /* in unit of byte */
	u32 user_virtual;
	u32 cmd_size;	    /* in unit of dword */
};

static int compat_via_chrome9_branch_buf_request(struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	struct umd_branch_buf_rec32 umd32;
	struct umd_branch_buf_rec __user *p_umd;
	int err;

	if (copy_from_user(&umd32, (void __user *)arg,
			sizeof(umd32)))
		return -EFAULT;

	p_umd = compat_alloc_user_space(sizeof(*p_umd));

	if (!access_ok(VERIFY_WRITE, p_umd, sizeof(*p_umd)))
		return -EFAULT;

	err = drm_ioctl(filp->f_path.dentry->d_inode, filp,
			DRM_IOCTL_VIA_CHROME9_REQUEST_BRANCH_BUF,
			(unsigned long)p_umd);

	if (err)
		return err;

	if (__get_user(umd32.id, &p_umd->id)
		|| __get_user(umd32.status, &p_umd->status)
		|| __get_user(umd32.offset, &p_umd->offset)
		|| __get_user(umd32.size, &p_umd->size))
		return -EFAULT;
#if CHROME9_COMPAT_DEBUG
	printk(KERN_ALERT "sheldon compat"" branch buffer request success!"
			" size=0x%x \n", umd32.size);
#endif
	if (copy_to_user((void __user *)arg, &umd32, sizeof(umd32)))
		return -EFAULT;
	return 0;
}

static int compat_via_chrome9_branch_buf_flush(struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	struct umd_branch_buf_rec32 umd32;
	struct umd_branch_buf_rec __user *p_umd;
	int err;

	if (copy_from_user(&umd32, (void __user *)arg,
			sizeof(umd32)))
		return -EFAULT;

	p_umd = compat_alloc_user_space(sizeof(*p_umd));

	if (!access_ok(VERIFY_WRITE, p_umd, sizeof(*p_umd))
		|| __put_user(umd32.id, &p_umd->id)
		|| __put_user(umd32.cmd_size, &p_umd->cmd_size))
		return -EFAULT;
#if CHROME9_COMPAT_DEBUG
	printk(KERN_ALERT "[CHROME9_COMPAT_DUBUG]: umk dump\n");
	int i = 0;
	u32 *p = (u32 *)&umd32;
	for(i=0; i < sizeof(umd32)/4; i++)
	{
		printk(KERN_ALERT "0x%08x \n",p[i]);
	}
#endif
#if CHROME9_COMPAT_DEBUG
	printk(KERN_ALERT "[CHROME9_COMPAT_DUBUG]"
			" go into compat_via_chrome9_branch_buf_flush,"
			" cmd_size = 0x%x  \n",umd32.cmd_size);
#endif
	err = drm_ioctl(filp->f_path.dentry->d_inode, filp,
			DRM_IOCTL_VIA_CHROME9_BRANCH_BUF_FLUSH,
			(unsigned long) p_umd);
	if (err)
		return err;
#if CHROME9_COMPAT_DEBUG
	printk(KERN_ALERT "[CHROME9_COMPAT_DEBUG]"
		"compat_via_chrome9_branch_buf_flush succfully called! \n");
#endif
	return 0;
}

drm_ioctl_compat_t *via_chrome9_compat_ioctls[] = {
	[DRM_VIA_CHROME9_INIT] = compat_via_chrome9_init,
	[DRM_VIA_CHROME9_FLUSH] = compat_via_chrome9_ioctl_flush,
	[DRM_VIA_CHROME9_ALLOCATE_VIDEO_MEM] =
		compat_via_chrome9_ioctl_allocate_mem_wrapper,
	[DRM_VIA_CHROME9_FREE_VIDEO_MEM] =
		compat_via_chrome9_ioctl_free_mem_wrapper,
	[DRM_VIA_CHROME9_ALLOCMEM] = compat_via_chrome9_ioctl_allocate_mem_base,
	[DRM_VIA_CHROME9_FREEMEM] = compat_via_chrome9_ioctl_freemem_base,
	[DRM_VIA_CHROME9_REQUEST_BRANCH_BUF] =
		compat_via_chrome9_branch_buf_request,
	[DRM_VIA_CHROME9_BRANCH_BUF_FLUSH] =
		compat_via_chrome9_branch_buf_flush
};


long via_chrome9_compat_ioctl(struct file *filp, unsigned int cmd,
		unsigned long arg)
{
	unsigned int nr = DRM_IOCTL_NR(cmd);
	drm_ioctl_compat_t *fn = NULL;
	int ret;

	if (nr < DRM_COMMAND_BASE)
		return drm_compat_ioctl(filp, cmd, arg);

	if (nr < DRM_COMMAND_BASE + DRM_ARRAY_SIZE(via_chrome9_compat_ioctls))
		fn = via_chrome9_compat_ioctls[nr - DRM_COMMAND_BASE];

	lock_kernel();
	if (fn != NULL)
		ret = (*fn) (filp, cmd, arg);
	else
		ret = drm_ioctl(filp->f_path.dentry->d_inode, filp, cmd, arg);
	unlock_kernel();

	return ret;
}

